@@ -46,33 +46,49 @@
 
 #define BOX_RADIOUS 10
 
+#define SCROLL_PERIOD 100 /* ms */
+#define SCROLL_STEP 1 /* pixel */
+
 struct _ELHomeAppletPrivate
 {
         RTComEl *eventlogger;
 
         GtkWidget *sender;
-        GtkWidget *message;
         GtkWidget *icon;
         GtkWidget *unread;
         GtkWidget *received;
         GtkWidget *empty;
+        GtkWidget *cut_message;
 
-        gint       event_id;
+        gchar *message;
+        gint event_id;
 
-        gboolean   active;
+        gboolean active;
 
         guint unread_count;
 
-        const gchar *current_font;
+        struct {
+                float red;
+                float green;
+                float blue;
+        } active_color;
+        PangoFontDescription *font_desc;
 
         guint idle_id;
+
+        cairo_surface_t *message_surface;
+
+        gboolean scroll_on_click;
+        gint scroll_offset;
+        gint hidden_message_height;
+        guint scroll_anim_id;
 };
 
 HD_DEFINE_PLUGIN_MODULE (ELHomeApplet, el_home_applet, HD_TYPE_HOME_PLUGIN_ITEM);
 
-const gchar* g_module_check_init(GModule *module);
+const gchar* g_module_check_init (GModule *module);
 const gchar*
-g_module_check_init(GModule *module)
+g_module_check_init (GModule *module)
 {
 	g_module_make_resident (module);
 	return NULL;
@@ -128,20 +144,123 @@ rounded_rectangle (cairo_t *cr,
                         x + r, y);
 }
 
+static cairo_surface_t*
+draw_text (cairo_t              *cr,
+           PangoFontDescription *desc,
+           const gchar          *text,
+           gint                  width,
+           gint                 *height)
+{
+        PangoLayout *layout;
+        PangoRectangle extent;
+
+        cairo_surface_t *gdk_surface, *result_surface;
+        cairo_t *msg_cr;
+
+        /* Create a PangoLayout, set the font and text */
+        layout = pango_cairo_create_layout (cr);
+        pango_layout_set_text (layout,
+                               text,
+                               -1);
+        pango_layout_set_font_description (layout, desc);
+
+        pango_layout_set_wrap (layout, PANGO_WRAP_WORD_CHAR);
+        pango_layout_set_width (layout, PANGO_SCALE*width);
+
+        pango_layout_get_pixel_extents (layout, NULL, &extent);
+        *height = extent.height;
+
+        gdk_surface = cairo_get_target (cr);
+        result_surface = cairo_surface_create_similar
+                (gdk_surface,
+                 CAIRO_CONTENT_COLOR_ALPHA,
+                 width,
+                 extent.height);
+        msg_cr = cairo_create (result_surface);
+
+        pango_cairo_update_layout (msg_cr, layout);
+        /* draw shadow */
+        cairo_move_to (msg_cr, 1, 1);
+        cairo_set_source_rgba (msg_cr, 0.2, 0.2, 0.2, 0.8);
+        pango_cairo_show_layout (msg_cr, layout);
+
+        /* draw fg */
+        cairo_move_to (msg_cr, 0, 0);
+        cairo_set_source_rgba (msg_cr, 1.0, 1.0, 1.0, 1.0);
+        pango_cairo_show_layout (msg_cr, layout);
+
+        cairo_destroy (msg_cr);
+        g_object_unref (layout);
+
+        return result_surface;
+}
+
+static void
+stop_scroll_anim (ELHomeAppletPrivate *priv)
+{
+        if (priv->scroll_anim_id > 0) {
+                g_source_remove (priv->scroll_anim_id);
+                priv->scroll_anim_id = 0;
+                priv->scroll_on_click = FALSE;
+                gtk_widget_hide (priv->cut_message);
+        }
+}
+
+static void
+style_set_cb (GtkWidget *widget,
+              GtkStyle  *previous_style,
+              ELHomeApplet *self)
+{
+        ELHomeAppletPrivate *priv = self->priv;
+        GdkColor color;
+        GtkStyle *font_style;
+
+        font_style = gtk_rc_get_style_by_paths (gtk_widget_get_settings (widget),
+                                                "SystemFont",
+                                                NULL,
+                                                G_TYPE_NONE);
+        if (font_style && font_style->font_desc) {
+                if (priv->font_desc)
+                        pango_font_description_free (priv->font_desc);
+                priv->font_desc = pango_font_description_copy (font_style->font_desc);
+        }
+
+        if (gtk_style_lookup_color (widget->style,
+                                    "ActiveTextColor",
+                                    &color)) {
+                priv->active_color.red = color.red/(float)G_MAXUINT16;
+                priv->active_color.green = color.green/(float)G_MAXUINT16;
+                priv->active_color.blue = color.blue/(float)G_MAXUINT16;
+        }
+}
+
+static void
+notify_on_current_desktop (GObject      *object,
+                           GParamSpec   *unused G_GNUC_UNUSED,
+                           ELHomeApplet *self)
+{
+        ELHomeAppletPrivate *priv = self->priv;
+        gboolean on;
+
+        g_object_get (object, "is-on-current-desktop", &on, NULL);
+        if (!on) {
+                stop_scroll_anim (self->priv);
+                priv->scroll_on_click = priv->scroll_offset;
+                priv->scroll_offset = 0;
+                if (priv->scroll_on_click)
+                        gtk_widget_show (priv->cut_message);
+                gtk_widget_queue_draw (GTK_WIDGET (self));
+        }
+}
+
 static gboolean
 expose_event (GtkWidget *self, GdkEventExpose *event)
 {
         ELHomeAppletPrivate *priv = EL_HOME_APPLET(self)->priv;
-
         cairo_t *cr;
-        GdkColor color;
-        float red, green, blue;
+        int message_height;
 
-        /* find theme active color */
-        gtk_style_lookup_color (self->style, "ActiveTextColor", &color);
-        red = color.red/(float)G_MAXUINT16;
-        green = color.green/(float)G_MAXUINT16;
-        blue = color.blue/(float)G_MAXUINT16;
+        message_height = C_HEIGHT - priv->received->allocation.height - HEADER_HEIGHT;
 
         cr = gdk_cairo_create (self->window);
         gdk_cairo_region (cr, event->region);
@@ -160,7 +279,7 @@ expose_event (GtkWidget *self, GdkEventExpose *event)
                            BOX_HEIGHT - 2*C_Y,
                            BOX_RADIOUS);
 
-        cairo_close_path(cr);
+        cairo_close_path (cr);
         cairo_stroke (cr);
 
         /* draw header */
@@ -175,11 +294,15 @@ expose_event (GtkWidget *self, GdkEventExpose *event)
         cairo_line_to (cr, C_WIDTH, HEADER_HEIGHT);
         cairo_line_to (cr, 0, HEADER_HEIGHT);
 
-        cairo_close_path(cr);
+        cairo_close_path (cr);
 
         cairo_set_source_rgba (cr, 0.2f, 0.2f, 0.2f, 0.8f);
         cairo_fill_preserve (cr);
-        cairo_set_source_rgba (cr, red, green, blue, 1.0f);
+        cairo_set_source_rgba (cr,
+                               priv->active_color.red,
+                               priv->active_color.green,
+                               priv->active_color.blue,
+                               1.0f);
         cairo_stroke (cr);
 
         /* draw body */
@@ -190,18 +313,26 @@ expose_event (GtkWidget *self, GdkEventExpose *event)
         cairo_curve_to (cr, C_WIDTH, C_HEIGHT, C_WIDTH, C_HEIGHT, C_WIDTH, C_HEIGHT - BOX_RADIOUS);
         cairo_line_to (cr, C_WIDTH, HEADER_HEIGHT);
         cairo_line_to (cr, 0, HEADER_HEIGHT);
-        cairo_close_path(cr);
+        cairo_close_path (cr);
 
         /* draw body filling depending on (in)active state */
         cairo_pattern_t *grad;
-        grad = cairo_pattern_create_linear(0, HEADER_HEIGHT,
-                                           0, C_HEIGHT);
+        grad = cairo_pattern_create_linear (0, HEADER_HEIGHT,
+                                            0, C_HEIGHT);
 
-        if (priv->active){
-                cairo_pattern_add_color_stop_rgba (grad, 0.5f,
-                                                   red, green, blue, 0.8f);
-                cairo_pattern_add_color_stop_rgba (grad, 1.0f,
-                                                   red/2, green/2, blue/2, 0.8f);
+        if (priv->active) {
+                cairo_pattern_add_color_stop_rgba (grad,
+                                                   0.5f,
+                                                   priv->active_color.red,
+                                                   priv->active_color.green,
+                                                   priv->active_color.blue,
+                                                   0.8f);
+                cairo_pattern_add_color_stop_rgba (grad,
+                                                   1.0f,
+                                                   priv->active_color.red/2,
+                                                   priv->active_color.green/2,
+                                                   priv->active_color.blue/2,
+                                                   0.8f);
         }
         else {
                 cairo_pattern_add_color_stop_rgba (grad, 0.5f,
@@ -220,8 +351,39 @@ expose_event (GtkWidget *self, GdkEventExpose *event)
         /*                    BOX_WIDTH - 2*C_X, */
         /*                    BOX_HEIGHT - 2*C_Y, */
         /*                    BOX_RADIOUS); */
-        /* cairo_close_path(cr); */
+        /* cairo_close_path (cr); */
         /* cairo_stroke (cr); */
+
+        /* draw message */
+        if (!priv->message_surface) {
+                gint height;
+
+                priv->message_surface = draw_text (cr,
+                                                   priv->font_desc,
+                                                   priv->message,
+                                                   MESSAGE_WIDTH,
+                                                   &height);
+
+                priv->hidden_message_height = height - message_height;
+                priv->scroll_on_click = priv->hidden_message_height > 0;
+                if (priv->scroll_on_click)
+                        gtk_widget_show (priv->cut_message);
+        }
+
+        cairo_rectangle (cr,
+                         2*C_X,
+                         HEADER_HEIGHT,
+                         MESSAGE_WIDTH,
+                         message_height);
+        cairo_clip (cr);
+
+        cairo_set_source_surface (cr,
+                                  priv->message_surface,
+                                  2*C_X,
+                                  HEADER_HEIGHT - priv->scroll_offset);
+        cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
+        cairo_paint (cr);
+
         cairo_pattern_destroy (grad);
         cairo_destroy (cr);
 
@@ -233,15 +395,25 @@ dispose (GObject *self)
 {
         ELHomeAppletPrivate *priv = EL_HOME_APPLET(self)->priv;
 
-        if (priv->idle_id){
+        stop_scroll_anim (priv);
+        if (priv->idle_id) {
                 g_source_remove (priv->idle_id);
                 priv->idle_id = 0;
         }
-        if (priv->eventlogger){
+        if (priv->eventlogger) {
                 g_object_unref (priv->eventlogger);
                 priv->eventlogger = NULL;
         }
 
+        if (priv->message) {
+                g_free (priv->message);
+                priv->message = NULL;
+        }
+        if (priv->font_desc) {
+                pango_font_description_free (priv->font_desc);
+                priv->font_desc = NULL;
+        }
+
         G_OBJECT_CLASS (el_home_applet_parent_class)->dispose (self);
 }
 
@@ -280,14 +452,18 @@ show_event (ELHomeApplet *self, RTComElIter *it)
 {
         ELHomeAppletPrivate *priv = self->priv;
 
-        gchar *message = NULL;
         gchar *remote = NULL;
         gchar *received = NULL;
         const gchar *icon_name = NULL;
 
-        if (it && rtcom_el_iter_first (it)){
-                rtcom_el_iter_dup_string (it, "free-text", &message);
-                if (message){
+        if (priv->message) {
+                g_free (priv->message);
+                priv->message = NULL;
+        }
+
+        if (it && rtcom_el_iter_first (it)) {
+                rtcom_el_iter_dup_string (it, "free-text", &priv->message);
+                if (priv->message) {
                         const gchar *service;
                         time_t received_t;
 
@@ -295,7 +471,7 @@ show_event (ELHomeApplet *self, RTComElIter *it)
                         if (rtcom_el_iter_get_int (it, "start-time", (gint*)&received_t))
                                 received = format_time (received_t);
 
-                        if(!rtcom_el_iter_dup_string (it, "remote-name", &remote))
+                        if (!rtcom_el_iter_dup_string (it, "remote-name", &remote))
                                 rtcom_el_iter_dup_string (it, "remote-id", &remote);
                         service = rtcom_el_iter_get_service (it);
                         if (!g_strcmp0 (service, "RTCOM_EL_SERVICE_SMS"))
@@ -304,24 +480,19 @@ show_event (ELHomeApplet *self, RTComElIter *it)
                                 icon_name = "chat_unread_chat";
                 }
         }
-        else{
+        else {
                 priv->event_id = -1;
         }
 
-        if (message){
-                if (!GTK_WIDGET_VISIBLE (priv->message))
-                        gtk_widget_show (priv->message);
-                gtk_label_set_text (GTK_LABEL (priv->message), message);
-        }
-        else{
-                gtk_widget_hide (priv->message);
+        if (priv->message)
+                gtk_widget_hide (priv->empty);
+        else
                 gtk_widget_show (priv->empty);
-        }
 
         gtk_label_set_text (GTK_LABEL (priv->sender), remote);
         gtk_label_set_text (GTK_LABEL (priv->received), received);
 
-        if (icon_name){
+        if (icon_name) {
                 const gchar *current_icon_name;
                 gtk_image_get_icon_name (GTK_IMAGE (priv->icon),
                                          &current_icon_name,
@@ -335,8 +506,17 @@ show_event (ELHomeApplet *self, RTComElIter *it)
         else
                 gtk_widget_hide (priv->icon);
 
-        g_free (message);
         g_free (remote);
+
+        stop_scroll_anim (priv);
+        priv->scroll_offset = 0;
+        if (priv->message_surface) {
+                cairo_surface_destroy (priv->message_surface);
+                priv->message_surface = NULL;
+        }
+
+        gtk_widget_hide (priv->cut_message);
+        gtk_widget_queue_draw (GTK_WIDGET (self));
 }
 
 static RTComElIter*
@@ -354,7 +534,7 @@ make_query (RTComEl *el, gint event_id)
 
         query = rtcom_el_query_new (el);
         rtcom_el_query_set_limit (query, 1);
-        if (event_id >= 0){
+        if (event_id >= 0) {
                 rtcom_el_query_prepare (query,
                                         "is-read", FALSE, RTCOM_EL_OP_EQUAL,
                                         "id", event_id, RTCOM_EL_OP_EQUAL,
@@ -362,7 +542,7 @@ make_query (RTComEl *el, gint event_id)
                                         "event-type", event_types, RTCOM_EL_OP_IN_STRV,
                                         NULL);
         }
-        else{
+        else {
                 rtcom_el_query_prepare (query,
                                         "is-read", FALSE, RTCOM_EL_OP_EQUAL,
                                         "service", services, RTCOM_EL_OP_IN_STRV,
@@ -370,7 +550,7 @@ make_query (RTComEl *el, gint event_id)
                                         NULL);
         }
         it = rtcom_el_get_events (el, query);
-        g_object_unref(query);
+        g_object_unref (query);
 
         return it;
 }
@@ -381,7 +561,7 @@ update_unread_label (ELHomeApplet *self)
         ELHomeAppletPrivate *priv = self->priv;
         gchar *text;
 
-        if (priv->unread_count > 0){
+        if (priv->unread_count > 0) {
                 text = g_strdup_printf ("%d", priv->unread_count);
                 gtk_label_set_text (GTK_LABEL (priv->unread), text);
                 g_free (text);
@@ -404,17 +584,17 @@ query_unread_events (RTComEl *el)
                                 "SELECT SUM(total_events)-SUM(read_events) FROM GroupCache;",
                                 -1,
                                 &stmt,
-                                NULL) != SQLITE_OK){
+                                NULL) != SQLITE_OK) {
                 g_error ("%s: can't compile SQL", G_STRFUNC);
                 return -1;
         }
 
         while (SQLITE_BUSY == (ret = sqlite3_step (stmt)));
 
-        if (ret == SQLITE_ROW){
+        if (ret == SQLITE_ROW) {
                 count = sqlite3_column_int (stmt, 0);
         }
-        else{
+        else {
                 g_error ("%s: error while executing SQL", G_STRFUNC);
         }
 
@@ -439,7 +619,7 @@ mark_as_read (ELHomeApplet *self)
 {
         ELHomeAppletPrivate *priv = self->priv;
 
-        if (priv->event_id >= 0){
+        if (priv->event_id >= 0) {
                 rtcom_el_set_read_event (priv->eventlogger,
                                          priv->event_id,
                                          TRUE,
@@ -489,21 +669,25 @@ new_event_cb (RTComEl      *backend,
 }
 
 static gboolean
-button_release_event_cb (GtkWidget      *widget,
-                         GdkEventButton *event,
-                         ELHomeApplet   *self)
+scroll_anim_cb (ELHomeApplet *self)
 {
         ELHomeAppletPrivate *priv = self->priv;
+        gboolean to_continue;
 
-        if (priv->active){
-                priv->active = FALSE;
-                gtk_widget_queue_draw (widget);
-#ifndef DEBUG_LAYOUT
-                mark_as_read (self);
-#endif
+        priv->scroll_offset += SCROLL_STEP;
+        gtk_widget_queue_draw_area (GTK_WIDGET (self),
+                                    3*C_X,
+                                    HEADER_HEIGHT + C_Y,
+                                    MESSAGE_WIDTH,
+                                    C_HEIGHT - priv->received->allocation.height - HEADER_HEIGHT);
+
+        to_continue = priv->scroll_offset <= priv->hidden_message_height;
+        if (!to_continue) {
+                priv->scroll_anim_id = 0;
+                gtk_widget_hide (priv->cut_message);
         }
 
-        return TRUE;
+        return to_continue;
 }
 
 static gboolean
@@ -513,7 +697,7 @@ button_press_event_cb (GtkWidget      *widget,
 {
         ELHomeAppletPrivate *priv = self->priv;
 
-        if (priv->event_id > 0){
+        if (priv->event_id > 0) {
                 priv->active = TRUE;
                 gtk_widget_queue_draw (widget);
         }
@@ -521,6 +705,34 @@ button_press_event_cb (GtkWidget      *widget,
         return TRUE;
 }
 
+static gboolean
+button_release_event_cb (GtkWidget      *widget,
+                         GdkEventButton *event,
+                         ELHomeApplet   *self)
+{
+        ELHomeAppletPrivate *priv = self->priv;
+
+        if (priv->active) {
+                priv->active = FALSE;
+                stop_scroll_anim (priv);
+                if (priv->scroll_on_click) {
+                        priv->scroll_on_click = FALSE;
+                        priv->scroll_anim_id = g_timeout_add (SCROLL_PERIOD,
+                                                              (GSourceFunc)scroll_anim_cb,
+                                                              self);
+                }
+                else {
+#ifndef DEBUG_LAYOUT
+                        mark_as_read (self);
+#endif
+                }
+
+                gtk_widget_queue_draw (widget);
+        }
+
+        return TRUE;
+}
+
 static gboolean
 leave_notify_event_cb (GtkWidget        *widget,
                        GdkEventCrossing *event,
@@ -528,8 +740,9 @@ leave_notify_event_cb (GtkWidget        *widget,
 {
         ELHomeAppletPrivate *priv = self->priv;
 
-        if (priv->active){
+        if (priv->active) {
                 priv->active = FALSE;
+                stop_scroll_anim (priv);
                 gtk_widget_queue_draw (widget);
         }
 
@@ -541,7 +754,7 @@ el_home_applet_init (ELHomeApplet *self)
 {
         ELHomeAppletPrivate *priv;
         GtkWidget *event_box;
-        GtkWidget *hbox, *vbox, *align;
+        GtkWidget *hbox, *vbox, *align, *footer;
 
         self->priv = EL_HOME_APPLET_GET_PRIVATE (self);
         priv = self->priv;
@@ -575,19 +788,7 @@ el_home_applet_init (ELHomeApplet *self)
         gtk_widget_set_name (priv->sender, "hildon-shadow-label");
         hildon_helper_set_logical_font (priv->sender, "SystemFont");
 
-        priv->message = g_object_new (GTK_TYPE_LABEL,
-                                      "label", "asdf asdf adsf asdf asdf asdf asdf asdf",
-                                      "wrap", TRUE,
-                                      "wrap-mode", PANGO_WRAP_WORD_CHAR,
-                                      NULL);
-
-        gtk_misc_set_alignment (GTK_MISC (priv->message),
-                                0.0f,
-                                0.0f);
-        gtk_widget_set_size_request (priv->message,
-                                     MESSAGE_WIDTH,
-                                     MESSAGE_HEIGHT);
-        gtk_widget_set_name (priv->message, "hildon-shadow-label");
+        priv->message = g_strdup ("One two three four five six seven eight nine ten");
 
         /* TODO: l10n */
         priv->empty = gtk_label_new ("No new messages");
@@ -598,22 +799,31 @@ el_home_applet_init (ELHomeApplet *self)
         gtk_misc_set_alignment (GTK_MISC (priv->received),
                                 1.0f,
                                 0.5f);
-        gtk_widget_set_size_request (priv->received,
-                                     MESSAGE_WIDTH,
-                                     -1);
         hildon_helper_set_logical_font (priv->received, "SmallSystemFont");
         gtk_widget_set_name (priv->received, "hildon-shadow-label");
 
+
+        priv->cut_message = gtk_label_new ("...");
+        gtk_misc_set_alignment (GTK_MISC (priv->cut_message),
+                                0.5f,
+                                0.0f);
+        hildon_helper_set_logical_font (priv->cut_message, "SmallSystemFont");
+        gtk_widget_set_name (priv->cut_message, "hildon-shadow-label");
+        GTK_WIDGET_SET_FLAGS (priv->cut_message, GTK_NO_SHOW_ALL);
+
         hbox = gtk_hbox_new (FALSE, 0);
         gtk_box_pack_start (GTK_BOX (hbox), priv->unread, FALSE, FALSE, 0);
         gtk_box_pack_start (GTK_BOX (hbox), priv->icon, FALSE, FALSE, 0);
         gtk_box_pack_start (GTK_BOX (hbox), priv->sender, TRUE, TRUE, 0);
 
+        footer = gtk_hbox_new (FALSE, 0);
+        gtk_box_pack_start (GTK_BOX (footer), priv->cut_message, TRUE, TRUE, 0);
+        gtk_box_pack_end (GTK_BOX (footer), priv->received, FALSE, FALSE, 0);
+
         vbox = gtk_vbox_new (FALSE, 0);
         gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
-        gtk_box_pack_start (GTK_BOX (vbox), priv->message, TRUE, TRUE, 0);
         gtk_box_pack_start (GTK_BOX (vbox), priv->empty, TRUE, TRUE, 0);
-        gtk_box_pack_start (GTK_BOX (vbox), priv->received, FALSE, FALSE, 0);
+        gtk_box_pack_end (GTK_BOX (vbox), footer, FALSE, FALSE, 0);
 
         align = gtk_alignment_new (0.5f, 0.0f, 1.0f, 1.0f);
         gtk_alignment_set_padding (GTK_ALIGNMENT (align),
@@ -629,12 +839,27 @@ el_home_applet_init (ELHomeApplet *self)
         gtk_container_add (GTK_CONTAINER (event_box), align);
         gtk_container_add (GTK_CONTAINER (self), event_box);
 
-        g_signal_connect (event_box, "button-press-event",
-                G_CALLBACK (button_press_event_cb), self);
-        g_signal_connect (event_box, "button-release-event",
-                G_CALLBACK (button_release_event_cb), self);
-        g_signal_connect (event_box, "leave-notify-event",
-                G_CALLBACK (leave_notify_event_cb), self);
+        g_signal_connect (event_box,
+                          "button-press-event",
+                          G_CALLBACK (button_press_event_cb),
+                          self);
+        g_signal_connect (event_box,
+                          "button-release-event",
+                          G_CALLBACK (button_release_event_cb),
+                          self);
+        g_signal_connect (event_box,
+                          "leave-notify-event",
+                          G_CALLBACK (leave_notify_event_cb),
+                          self);
+
+        g_signal_connect (event_box,
+                          "style-set",
+                          G_CALLBACK (style_set_cb),
+                          self);
+        g_signal_connect (self,
+                          "notify::is-on-current-desktop",
+                          G_CALLBACK (notify_on_current_desktop),
+                          self);
 
         gtk_widget_show_all (GTK_WIDGET (event_box));
 
